module SSG.HTML

import public SSG.HTML.ElementTypes

import Data.String
import Decidable.Equality

import public Structures.Dependent.DList

export infix 8 =$

||| An attribute within an HTML tag
public export
data Attribute : Type where
  -- TODO: Add some structure here, maybe make `=$` a smart constructor
  (=$) : (attribute, value : String) -> Attribute

namespace Attribute
  ||| Convert an attribute tag into a string suitable for rendering a tag with
  export
  toString : Attribute -> String
  toString (attribute =$ value) =
    let value_cs = unpack value
    in if any (== '\'') value_cs
      then "\{attribute}=\"\{value}\""
      else if any (== '"') value_cs || any (== ' ') value_cs
        then "\{attribute}='\{value}'"
        else "\{attribute}=\{value}"

||| An element in an HTML document, indexed by the tag type.
|||
||| Raw text has a tag type of `""`
public export
data Html : String -> Type where
  ||| Textual content outside of a tag
  Text : (content : String)
    -> Html ""
  ||| A void type tag, has attributes, no contents
  Void : (type : String) -> {auto void_prf : IsVoid type}
    -> (attributes : List Attribute)
    -> Html type
  ||| A normal type tag, has attributes, and contents
  Normal : (type : String) -> {auto normal_prf : IsNormal type}
    -> (attributes : List Attribute)
    -> {content_types : _} -> (contents : DList _ Html content_types)
    -> Html type
  ||| A raw text or an escapable raw text element
  RawText : (type : String)
    -> {auto raw_prf : Either (IsRawText type) (IsEscapableRawText type)}
    -> (attributes : List Attribute)
    -> (content : String)
    -> Html type

namespace Html
  ||| Convert to a "pretty printed" string containing the html, starting at the
  ||| specified indentation level
  export
  viewIndented : (indent_level : Nat) -> Html type -> String
  viewIndented indent_level (Text content) =
    replicate (indent_level * 2) ' ' ++ content
  viewIndented indent_level (Void type attributes) =
    let indent = replicate (indent_level * 2) ' ' in
    if length attributes > 0
      then
        let attrs = joinBy " " $ map toString attributes
        in "<\{type} \{attrs} />"
      else "<\{type} />"
  viewIndented indent_level (Normal type attributes {content_types} contents) =
    let indent = replicate (indent_level * 2) ' ' in
    -- Special handling if the tag contains exactly one `Text` element, we won't
    -- do any newlines between the tag and its contents in that case
    case decEq [""] content_types of
      Yes prf => assert_total $ -- BUG Compiler Bug ???
        let [Text content] : DList _ Html [""] = rewrite prf in contents
        in if length attributes > 0
          then
            let attrs = joinBy " " $ map toString attributes
            in "\{indent}<\{type} \{attrs}>\{content}</\{type}>"
          else "\{indent}<\{type}>\{content}</\{type}>"
      No contra =>
        let end = "\n\{indent}</\{type}>"
            inner = joinBy "\n" $
                      dMap (\_, x => viewIndented (S indent_level) x) contents
        in if length attributes > 0
          then
            let attrs = joinBy " " $ map toString attributes
            in "\{indent}<\{type} \{attrs}>\n" ++ inner ++ end
          else "\{indent}<\{type}>\n" ++ inner ++ end
  viewIndented indent_level (RawText type attributes content) =
    let indent = replicate (indent_level * 2) ' ' in
    if length attributes > 0
      then
        let attrs = joinBy " " $ map toString attributes
        in "\{indent}<\{type} \{attrs}>\{content}</\{type}>"
      else "\{indent}<\{type}>\{content}</\{type}>"

  ||| Convert to a "pretty printed" string containing the html
  export
  view : Html type -> String
  view = viewIndented 0

  ||| Render a top level html document, including the doctype tag
  export
  render : Html "html" -> String
  render x =
    "<!DOCTYPE HTML>\n" ++ viewIndented 0 x

  ||| Cacluate the type signature of the `tag` function
  public export
  TagType : IsValidTag type -> Type
  TagType (PVoid type) =
    (attributes : List Attribute) -> Html type
  TagType (PRawText type) =
    (attributes : List Attribute) -> (content : String) -> Html type
  TagType (PEscapableRawText type) =
    (attributes : List Attribute) -> (content : String) -> Html type
  TagType (PNormal type) =
    (attributes : List Attribute)
    -> {content_types : _} -> (contents : DList _ Html content_types)
    -> Html type

  public export
  TagType' : (type : String) -> {auto prf : IsValidTag type} -> Type
  TagType' type {prf} = TagType prf

  -- ||| Smart constructor for tags
  -- public export
  -- tag : (type : String) -> {auto prf : IsValidTag type} -> TagType prf
  -- tag type {prf = (PVoid type)} = Void type
  -- tag type {prf = (PRawText type)} = RawText type
  -- tag type {prf = (PEscapableRawText type)} = RawText type
  -- tag type {prf = (PNormal type)} = Normal type
